import logging
import sys
import orjson
import threading
import random
import os
import zmq
import zmq.auth
import signal
import importlib.metadata

from zmq.auth.thread import ThreadAuthenticator
from multiprocessing import Event
from typing import Union, Any, List, Optional
from . import NFP
from .inventory import NorFabInventory, logging_config_producer
from .keepalives import KeepAliver
from .security import generate_certificates

log = logging.getLogger(__name__)

try:
    signal.signal(signal.SIGINT, signal.SIG_IGN)
except Exception:
    pass

# ----------------------------------------------------------------------
# NORFAB Protocol Broker Implementation
# ----------------------------------------------------------------------


class NFPService(object):
    """A single NFP Service"""

    __slots__ = ("name", "workers")

    def __init__(self, name: str):
        self.name = name  # Service name
        self.workers = []  # list of known workers


class NFPWorker(object):
    """
    An NFP Worker convenience class.

    Attributes:
        service (NFPService): The service instance.
        ready (bool): Indicates if the worker is ready.
        exit_event (threading.Event): Event to signal exit.
        keepalive (int): Keepalive interval in milliseconds.
        multiplier (int): Multiplier value.

    Methods:
        start_keepalives():
            Starts the keepalive process for the worker.
        is_ready() -> bool:
            Checks if the worker has signaled W.READY.
        destroy(disconnect=False):
            Cleans up the worker, optionally disconnecting it.

    Args:
        address (str): Address to route to.
        socket: The socket object used for communication.
        socket_lock: The lock object to synchronize socket access.
        multiplier (int): Multiplier value, e.g., 6 times.
        keepalive (int): Keepalive interval in milliseconds, e.g., 5000 ms.
        service (NFPService, optional): The service instance. Defaults to None.
    """

    def __init__(
        self,
        address: str,
        socket: Any,
        socket_lock: Any,
        multiplier: int,  # e.g. 6 times
        keepalive: int,  # e.g. 5000 ms
        service: Optional[NFPService] = None,
    ):
        self.address = address  # Address to route to
        self.service = service
        self.ready = False
        self.socket = socket
        self.exit_event = threading.Event()
        self.keepalive = keepalive
        self.multiplier = multiplier
        self.socket_lock = socket_lock
        self.build_message = NFP.MessageBuilder()

    def start_keepalives(self) -> None:
        self.keepaliver = KeepAliver(
            address=self.address,
            socket=self.socket,
            multiplier=self.multiplier,
            keepalive=self.keepalive,
            exit_event=self.exit_event,
            service=self.service.name,
            whoami=NFP.BROKER,
            name="NFPBroker",
            socket_lock=self.socket_lock,
        )
        self.keepaliver.start()

    def is_ready(self) -> bool:
        """
        Check if the worker is ready.

        Returns:
            bool: True if the worker has signaled readiness (W.READY) and the service is not None, otherwise False.
        """
        return self.service is not None and self.ready is True

    def destroy(self, disconnect: bool = False) -> None:
        """
        Clean up routine for the worker.

        This method performs the following actions:

        1. Sets the exit event to signal termination.
        2. Stops the keepaliver if it exists.
        3. Removes the current worker from the service's worker list.
        4. Optionally sends a disconnect message to the broker if `disconnect` is True.

        Args:
            disconnect (bool): If True, sends a disconnect message to the broker.
        """
        self.exit_event.set()
        if hasattr(self, "keepaliver"):
            self.keepaliver.stop()
        if self.service is not None:
            self.service.workers.remove(self)

        if disconnect is True and self.service is not None:
            msg = self.build_message.broker_to_worker_disconnect(
                worker_address=self.address, service=self.service.name
            )
            with self.socket_lock:
                self.socket.send_multipart(msg)


class NFPBroker:
    """
    Attributes:
        private_keys_dir (str): Directory for private keys.
        public_keys_dir (str): Directory for public keys.
        broker_private_key_file (str): File path for broker's private key.
        broker_public_key_file (str): File path for broker's public key.
        keepalive (int): The keepalive interval.
        multiplier (int): The multiplier value.
        services (dict): A dictionary to store services.
        workers (dict): A dictionary to store workers.
        exit_event (Event): The event to signal the broker to exit.
        inventory (NorFabInventory): The inventory object.
        base_dir (str): The base directory path from the inventory.
        broker_base_dir (str): The broker's base directory path.
        ctx (zmq.Context): The ZeroMQ context.
        auth (ThreadAuthenticator): The authenticator for the ZeroMQ context.
        socket (zmq.Socket): The ZeroMQ socket.
        poller (zmq.Poller): The ZeroMQ poller.
        socket_lock (threading.Lock): The lock to protect the socket object.

    Methods:
        setup_logging(self, log_queue, log_level: str) -> None:
            Method to apply logging configuration.
        mediate(self):
            Main broker work happens here.
        destroy(self):
            Disconnect all workers, destroy context.
        delete_worker(self, worker, disconnect):
            Deletes worker from all data structures, and deletes worker.
        purge_workers(self):
            Look for & delete expired workers.
        send_to_worker(self, worker: NFPWorker, command: bytes, sender: bytes, uuid: bytes, data: bytes):
            Send message to worker. If message is provided, sends that message.
        send_to_client(self, client: str, command: str, service: str, message: list):
            Send message to client.
        process_worker(self, sender, msg):
            Process message received from worker.
        require_worker(self, address):
            Finds the worker, creates if necessary.
        require_service(self, name):
            Locates the service (creates if necessary).
        process_client(self, sender, msg):
            Process a request coming from a client.
        filter_workers(self, target: bytes, service: NFPService) -> list:
            Helper function to filter workers.
        dispatch(self, sender, command, service, target, uuid, data):
            Dispatch requests to waiting workers as possible.
        mmi_service(self, sender, command, target, uuid, data):
            Handle internal service according to 8/MMI specification.
        inventory_service(self, sender, command, target, uuid, data):
            Handle inventory service requests.

    Args:
        endpoint (str): The endpoint address for the broker to bind to.
        exit_event (Event): An event to signal the broker to exit.
        inventory (NorFabInventory): The inventory object containing configuration and state.
        log_level (str, optional): The logging level. Defaults to None.
        log_queue (object, optional): The logging queue. Defaults to None.
        multiplier (int, optional): A multiplier value for internal use. Defaults to 6.
        keepalive (int, optional): The keepalive interval in milliseconds. Defaults to 2500.
        init_done_event (Event, optional): An event to signal that initialization is done. Defaults to None.
    """

    private_keys_dir = None
    public_keys_dir = None
    broker_private_key_file = None
    broker_public_key_file = None

    def __init__(
        self,
        endpoint: str,
        exit_event: Event,
        inventory: NorFabInventory,
        log_level: str = None,
        log_queue: object = None,
        multiplier: int = 6,
        keepalive: int = 2500,
        init_done_event: Event = None,
    ):
        self.setup_logging(log_queue, log_level)
        self.keepalive = keepalive
        self.multiplier = multiplier
        init_done_event = init_done_event or Event()

        self.services = {}
        self.workers = {}
        self.build_message = NFP.MessageBuilder()
        self.exit_event = exit_event
        self.inventory = inventory
        self.zmq_auth = self.inventory.broker.get("zmq_auth", True)

        self.base_dir = self.inventory.base_dir
        self.broker_base_dir = os.path.join(
            self.base_dir, "__norfab__", "files", "broker"
        )
        os.makedirs(self.base_dir, exist_ok=True)
        os.makedirs(self.broker_base_dir, exist_ok=True)

        self.ctx = zmq.Context()
        self.socket = self.ctx.socket(zmq.ROUTER)
        self.socket.linger = (
            0  # discard pending messages immediately when the socket is closed
        )
        self.socket.setsockopt(
            zmq.ROUTER_HANDOVER, 1
        )  # allow worker/client reconnect using same identity

        # generate certificates, create directories and load certs
        if self.zmq_auth is not False:
            generate_certificates(
                self.broker_base_dir, cert_name="broker", inventory=inventory
            )
            self.private_keys_dir = os.path.join(self.broker_base_dir, "private_keys")
            self.public_keys_dir = os.path.join(self.broker_base_dir, "public_keys")
            self.broker_private_key_file = os.path.join(
                self.private_keys_dir, "broker.key_secret"
            )
            self.broker_public_key_file = os.path.join(
                self.public_keys_dir, "broker.key"
            )
            server_public, server_secret = zmq.auth.load_certificate(
                self.broker_private_key_file
            )

            # Start an authenticator for this context.
            self.auth = ThreadAuthenticator(self.ctx)
            self.auth.start()
            # self.auth.allow("0.0.0.0")
            self.auth.allow_any = True
            # Tell the authenticator how to handle CURVE requests
            self.auth.configure_curve(location=zmq.auth.CURVE_ALLOW_ANY)
            self.socket.curve_secretkey = server_secret
            self.socket.curve_publickey = server_public
            self.socket.curve_server = True  # must come before bind

        self.poller = zmq.Poller()
        self.poller.register(self.socket, zmq.POLLIN)
        self.socket.bind(endpoint)
        self.socket_lock = (
            threading.Lock()
        )  # used for keepalives to protect socket object

        init_done_event.set()  # signal finished initializing broker
        log.debug(f"NFPBroker - is ready and listening on {endpoint}")

    def setup_logging(self, log_queue, log_level: str) -> None:
        """
        Configures logging for the application.

        This method sets up the logging configuration using a provided log queue and log level.
        It updates the logging configuration dictionary with the given log queue and log level,
        and then applies the configuration using `logging.config.dictConfig`.

        Args:
            log_queue (queue.Queue): The queue to be used for logging.
            log_level (str): The logging level to be set. If None, the default level is used.

        Returns:
            None
        """
        logging_config_producer["handlers"]["queue"]["queue"] = log_queue
        if log_level is not None:
            logging_config_producer["root"]["level"] = log_level
        logging.config.dictConfig(logging_config_producer)

    def mediate(self):
        """
        Main broker work happens here.

        This method continuously polls for incoming messages and processes them
        based on their headers. It handles messages from clients and workers,
        and purges inactive workers periodically. The method also checks for an
        exit event to gracefully shut down the broker.

        Raises:
            KeyboardInterrupt: If the process is interrupted by a keyboard signal.
        """
        while True:
            try:
                items = self.poller.poll(1000)
            except KeyboardInterrupt:
                break  # Interrupted

            if items:
                with self.socket_lock:
                    msg = self.socket.recv_multipart()
                log.debug(f"NFPBroker - received '{msg}'")

                if len(msg) < 3:
                    log.error(f"NFPBroker - received malformed message: {msg}")
                    continue

                sender = msg.pop(0)
                empty = msg.pop(0)  # noqa
                header = msg.pop(0)

                if header == NFP.CLIENT:
                    self.process_client(sender, msg)
                elif header == NFP.WORKER:
                    self.process_worker(sender, msg)
                else:
                    log.error(
                        f"NFPBroker - message from '{sender}' contains unsupported header '{header}'"
                    )

            self.purge_workers()

            # check if need to stop
            if self.exit_event.is_set():
                self.destroy()
                break

    def destroy(self):
        """
        Disconnect all workers and destroy the context.

        This method performs the following actions:

        1. Logs an interrupt message indicating that the broker is being killed.
        2. Iterates through all
        """
        log.info("NFPBroker - interrupt received, killing broker")
        for name in list(self.workers.keys()):
            # in case worker self destroyed while we iterating
            if self.workers.get(name):
                self.delete_worker(self.workers[name], True)
        if self.zmq_auth is not False:
            self.auth.stop()
        self.ctx.destroy(0)

    def delete_worker(self, worker: NFPWorker, disconnect: bool) -> None:
        """
        Deletes a worker from all data structures and destroys the worker.

        Args:
            worker (Worker): The worker instance to be deleted.
            disconnect (bool): A flag indicating whether to disconnect the worker before deletion.

        Returns:
            None
        """
        worker.destroy(disconnect)
        self.workers.pop(worker.address, None)

    def purge_workers(self) -> None:
        """
        Look for and delete expired workers.

        This method iterates through the list of workers and checks if each worker's
        keepalive thread is still alive. If a worker's keepalive thread is not alive,
        the worker is considered expired and is deleted from the list of workers.
        Additionally, a log message is generated indicating that the worker's keepalive
        has expired.

        Note:
            The method handles the case where a worker might be destroyed while
            iterating through the list of workers.

        Logging:
            Logs an info message when a worker's keepalive has expired, including the
            worker's address.
        """
        for name in list(self.workers.keys()):
            # in case worker self destroyed while we iterating
            if self.workers.get(name):
                w = self.workers[name]
                if not w.keepaliver.is_alive():
                    self.delete_worker(w, False)
                    log.info(
                        f"NFPBroker - {w.address.decode(encoding='utf-8')} worker keepalives expired"
                    )

    def send_to_worker(
        self, worker: NFPWorker, command: bytes, sender: bytes, uuid: bytes, data: bytes
    ) -> None:
        """
        Send a message to a worker. If a message is provided, sends that message.

        Args:
            worker (NFPWorker): The worker to send the message to.
            command (bytes): The command to send (e.g., NFP.POST or NFP.GET).
            sender (bytes): The sender's identifier.
            uuid (bytes): The unique identifier for the message.
            data (bytes): The data to be sent with the message.

        Logs:
            Logs an error if the command is invalid.
            Logs a debug message when sending the message to the worker.
        """
        # Stack routing and protocol envelopes to start of message
        if command == NFP.POST:
            msg = self.build_message.broker_to_worker_post(
                worker_address=worker.address, sender=sender, uuid=uuid, data=data
            )
        elif command == NFP.GET:
            msg = self.build_message.broker_to_worker_get(
                worker_address=worker.address, sender=sender, uuid=uuid, data=data
            )
        elif command == NFP.PUT:
            msg = self.build_message.broker_to_worker_put(
                worker_address=worker.address, sender=sender, uuid=uuid, data=data
            )
        elif command == NFP.MMI:
            msg = self.build_message.broker_to_worker_mmi(
                worker_address=worker.address, sender=sender, uuid=uuid, data=data
            )
        else:
            log.error(f"NFPBroker - invalid worker command: {command}")
            return
        with self.socket_lock:
            log.debug(
                f"NFPBroker - sending command '{command}' to worker '{worker.address}', job '{uuid}', from client '{sender}'"
            )
            self.socket.send_multipart(msg)

    def send_to_client(
        self, client: str, command: str, service: str, message: list
    ) -> None:
        """
        Send a message to a specified client.

        Args:
            client (str): The identifier of the client to send the message to.
            command (str): The command type, either 'RESPONSE' or 'EVENT'.
            service (str): The service associated with the message.
            message (list): The message content to be sent.
        """

        # Stack routing and protocol envelopes to start of message
        if command == NFP.RESPONSE:
            msg = self.build_message.broker_to_client_response(
                client=client, service=service, message=message
            )
        elif command == NFP.EVENT:
            msg = self.build_message.broker_to_client_event(
                client=client, service=service, message=message
            )
        elif command == NFP.STREAM:
            msg = self.build_message.broker_to_client_stream(
                client=client, service=service, message=message
            )
        elif command == NFP.MMI:
            msg = self.build_message.broker_to_client_mmi(
                client=client, service=service, message=message
            )
        else:
            log.error(f"NFPBroker - invalid client command: {command}")
            return
        with self.socket_lock:
            log.debug(
                f"NFPBroker - sending to client '{client}', command '{command}', service '{service}'"
            )
            self.socket.send_multipart(msg)

    def process_worker(self, sender: str, msg: list):
        """
        Process message received from worker.

        Parameters:
            sender (str): The identifier of the sender (worker).
            msg (list): The message received from the worker, where the first element is the command.

        Commands:

        - NFP.READY: Marks the worker as ready and assigns a service to it.
        - NFP.RESPONSE: Sends a response to a client.
        - NFP.KEEPALIVE: Processes a keepalive message from the worker.
        - NFP.DISCONNECT: Handles worker disconnection.
        - NFP.EVENT: Sends an event to a client.
        - NFP.STREAM: Sends an stream data to a client.

        If the worker is not ready and an invalid command is received, the worker is deleted.

        Raises:
            AttributeError: If the worker does not have the required attributes for certain commands.
            IndexError: If the message list does not contain the expected number of elements for certain commands.
        """
        command = msg.pop(0)
        worker = self.require_worker(sender)

        log.debug(f"NFPBroker - processing '{sender}' worker message: '{msg}'")

        if NFP.READY == command and not worker.is_ready():
            service = msg.pop(0)
            worker.service = self.require_service(service)
            worker.ready = True
            worker.start_keepalives()
            worker.service.workers.append(worker)
        elif NFP.RESPONSE == command and worker.is_ready():
            client = msg.pop(0)
            empty = msg.pop(0)
            self.send_to_client(client, NFP.RESPONSE, worker.service.name, msg)
        elif NFP.KEEPALIVE == command and hasattr(worker, "keepaliver"):
            worker.keepaliver.received_heartbeat([worker.address] + msg)
        elif NFP.DISCONNECT == command and worker.is_ready():
            self.delete_worker(worker, False)
        elif NFP.EVENT == command and worker.is_ready():
            client = msg.pop(0)
            empty = msg.pop(0)  # noqa
            self.send_to_client(client, NFP.EVENT, worker.service.name, msg)
        elif NFP.STREAM == command and worker.is_ready():
            client = msg.pop(0)
            empty = msg.pop(0)  # noqa
            self.send_to_client(client, NFP.STREAM, worker.service.name, msg)
        elif not worker.is_ready():
            self.delete_worker(worker, disconnect=True)
        else:
            log.error(
                f"NFPBroker - invalid message: {msg}, command: {command}, sender: {sender}"
            )

    def require_worker(self, address: bytes) -> NFPWorker:
        """
        Finds the worker associated with the given address, creating a new worker if necessary.

        Args:
            address (str): The address of the worker to find or create.

        Returns:
            NFPWorker: The worker associated with the given address.
        """
        if not self.workers.get(address):
            self.workers[address] = NFPWorker(
                address=address,
                socket=self.socket,
                multiplier=self.multiplier,
                keepalive=self.keepalive,
                socket_lock=self.socket_lock,
            )
            log.info(
                f"NFPBroker - registered new worker {address.decode(encoding='utf-8')}"
            )

        return self.workers[address]

    def require_service(self, name: bytes) -> NFPService:
        """
        Locates the service by name, creating it if necessary.

        Args:
            name (str): The name of the service to locate or create.

        Returns:
            NFPService: The located or newly created service instance.
        """
        if not self.services.get(name):
            service = NFPService(name)
            self.services[name] = service
            log.debug(
                f"NFPBroker - registered new service {name.decode(encoding='utf-8')}"
            )

        return self.services[name]

    def process_client(self, sender: str, msg: list) -> None:
        """
        Process a request coming from a client.

        Args:
            sender (str): The identifier of the client sending the request.
            msg (list): The message received from the client, expected to be a list where the first five elements are:
                - command (str): The command issued by the client.
                - service (str): The service to which the command is directed.
                - target (str): The target of the command.
                - uuid (str): The unique identifier for the request.
                - data (any): The data associated with the request.

        Raises:
            ValueError: If the command is not recognized as a valid client command.

        The method processes the command by:
            - Checking if the command is valid.
            - Routing the request to the appropriate service handler based on the service specified.
            - Sending an error message back to the client if the command is unsupported.
        """
        command = msg.pop(0)
        service = msg.pop(0)
        target = msg.pop(0)
        uuid = msg.pop(0)
        data = msg.pop(0)

        # check if valid command from client
        if command not in NFP.client_commands:
            message = f"NFPBroker - Unsupported client command '{command}'"
            log.error(message)
            self.send_to_client(
                sender, NFP.RESPONSE, service, [message.encode("utf-8")]
            )
        # Management Interface
        elif service == b"mmi.service.broker":
            self.mmi_service(sender, command, target, uuid, data)
        elif service == b"sid.service.broker":
            self.inventory_service(sender, command, target, uuid, data)
        else:
            self.dispatch(
                sender, command, self.require_service(service), target, uuid, data
            )

    def filter_workers(self, target: bytes, service: NFPService) -> list:
        """
        Helper function to filter workers

        Args:
            target: bytest string, workers target
            service: NFPService object
        """
        ret = []

        if service.name == b"all":
            if target == b"all":
                ret = self.workers.values()
            elif target == b"any":
                ret = [
                    s.workers[random.randint(0, len(s.workers) - 1)]
                    for s in self.services.values()
                    if s.workers
                ]
            elif target in self.workers:  # single worker
                ret = [self.workers[target]]
            else:  # target list of workers
                try:
                    target = orjson.loads(target)
                    if isinstance(target, list):
                        for w in target:
                            w = w.encode("utf-8")
                            if w in self.workers:
                                ret.append(self.workers[w])
                        ret = list(set(ret))  # dedup workers
                except Exception as e:
                    log.error(
                        f"NFPBroker - Failed to load target '{target}' with error '{e}'"
                    )
        elif not service.workers:
            log.warning(
                f"NFPBroker - '{service.name}' has no active workers registered, try later"
            )
            ret = []
        elif target == b"any":
            ret = [service.workers[random.randint(0, len(service.workers) - 1)]]
        elif target == b"all":
            ret = service.workers
        elif target in self.workers:  # single worker
            ret = [self.workers[target]]
        else:  # target list of workers
            try:
                target = orjson.loads(target)
                if isinstance(target, list):
                    for w in target:
                        w = w.encode("utf-8")
                        if w in self.workers:
                            ret.append(self.workers[w])
                    ret = list(set(ret))  # dedup workers
            except Exception as e:
                log.error(
                    f"NFPBroker - Failed to load target '{target}' with error '{e}'"
                )
        return ret

    def dispatch(
        self,
        sender: str,
        command: bytes,
        service: NFPService,
        target: Union[str, List[str]],
        uuid: str,
        data: Any,
    ) -> None:
        """
        Dispatch requests to waiting workers as possible

        Args:
            sender (str): The sender of the request.
            command (bytes): The command to be executed by the workers.
            service (Service): The service object associated with the request.
            target (str): A string indicating the addresses of the workers to dispatch to.
            uuid (str): A unique identifier for the request.
            data (Any): The data to be sent to the workers.
        """
        log.debug(
            f"NFPBroker - dispatching request to workers or clients: sender '{sender}', "
            f"command '{command}', service '{service.name}', target '{target}', "
            f"data '{data}', uuid '{uuid}'"
        )
        self.purge_workers()
        workers = self.filter_workers(target, service)

        # handle case when service has no workers registered
        if not workers:
            message = f"NFPBroker - {service.name} service failed to target workers '{target}'"
            log.error(message)
            self.send_to_client(
                sender,
                NFP.RESPONSE,
                service.name,
                [
                    uuid,
                    b"400",
                    orjson.dumps(
                        {
                            "workers": None,
                            "uuid": uuid.decode("utf-8"),
                            "target": target.decode("utf-8"),
                            "status": "FAILED",
                            "service": service.name.decode("utf-8"),
                            "errors": [message],
                        }
                    ),
                ],
            )
            return

        # inform client that JOB dispatched for POST / GET requests
        if command in [NFP.POST, NFP.GET]:
            w_addresses = [w.address.decode("utf-8") for w in workers]
            self.send_to_client(
                sender,
                NFP.RESPONSE,
                service.name,
                [
                    uuid,
                    b"202",
                    orjson.dumps(
                        {
                            "workers": w_addresses,
                            "uuid": uuid.decode("utf-8"),
                            "target": target.decode("utf-8"),
                            "status": "DISPATCHED",
                            "service": service.name.decode("utf-8"),
                        }
                    ),
                ],
            )

        # send job to workers
        for worker in workers:
            self.send_to_worker(worker, command, sender, uuid, data)

    def mmi_service(self, sender, command, target, uuid, data):
        """
        Handle internal broker Management Interface (MMI) service tasks.

        Parameters:
            sender (str): The sender of the request.
            command (str): The command to be executed.
            target (str): The target of the command.
            uuid (str): The unique identifier for the request.
            data (str): The data payload in JSON format.

        Supported MMI Tasks:

        - "show_workers": Returns a list of workers with their details.
        - "show_broker": Returns broker details including endpoint, status, keepalives, workers count, services count, directories, and security.
        - "show_broker_version": Returns the version of various packages and the platform.
        - "show_broker_inventory": Returns the broker's inventory.

        The response is sent back to the client in a format of JSON formatted string.
        """
        log.debug(
            f"mmi.service.broker - processing request: sender '{sender}', "
            f"command '{command}', target '{target}'"
            f"data '{data}', uuid '{uuid}'"
        )
        data = orjson.loads(data)
        task = data.get("task")
        kwargs = data.get("kwargs", {})
        ret = f"Unsupported task '{task}'"
        if task == "show_workers":
            if self.workers:
                ret = [
                    {
                        "name": w.address.decode("utf-8"),
                        "service": w.service.name.decode("utf-8"),
                        "status": "alive" if w.keepaliver.is_alive() else "dead",
                        "holdtime": str(w.keepaliver.show_holdtime()),
                        "keepalives tx/rx": f"{w.keepaliver.keepalives_send} / {w.keepaliver.keepalives_received}",
                        "alive (s)": str(w.keepaliver.show_alive_for()),
                    }
                    for k, w in self.workers.items()
                ]
                # filter reply
                service = kwargs.get("service")
                status = kwargs.get("status")
                if service and service != "all":
                    ret = [w for w in ret if w["service"] == service]
                if status in ["alive", "dead"]:
                    ret = [w for w in ret if w["status"] == status]
                if not ret:
                    ret = [{"name": "", "service": "", "status": ""}]
            else:
                ret = [{"name": "", "service": "", "status": ""}]
        elif task == "show_broker":
            ret = {
                "endpoint": self.socket.getsockopt_string(zmq.LAST_ENDPOINT),
                "status": "active",
                "keepalives": {
                    "interval": self.keepalive,
                    "multiplier": self.multiplier,
                },
                "workers count": len(self.workers),
                "services count": len(self.services),
                "directories": {
                    "base-dir": self.base_dir,
                    "private-keys-dir": self.private_keys_dir,
                    "public-keys-dir": self.public_keys_dir,
                },
                "security": {
                    "broker-private-key-file": self.broker_private_key_file,
                    "broker-public-key-file": self.broker_public_key_file,
                    "zmq-auth": self.zmq_auth,
                },
            }
        elif task == "show_broker_version":
            ret = {
                "norfab": "",
                "pyyaml": "",
                "pyzmq": "",
                "psutil": "",
                "pydantic": "",
                "tornado": "",
                "jinja2": "",
                "python": sys.version.split(" ")[0],
                "platform": sys.platform,
            }
            # get version of packages installed
            for pkg in ret.keys():
                try:
                    ret[pkg] = importlib.metadata.version(pkg)
                except importlib.metadata.PackageNotFoundError:
                    pass
        elif task == "show_broker_inventory":
            ret = self.inventory.dict()
        reply = orjson.dumps(ret)
        self.send_to_client(
            sender, NFP.RESPONSE, b"mmi.service.broker", [uuid, b"200", reply]
        )
        self.send_to_client(
            sender, NFP.MMI, b"mmi.service.broker", [uuid, b"200", reply]
        )

    def inventory_service(self, sender, command, target, uuid, data):
        log.debug(
            f"sid.service.broker - processing request: sender '{sender}', "
            f"command '{command}', target '{target}'"
            f"data '{data}', uuid '{uuid}'"
        )
        data = orjson.loads(data)
        task = data.get("task")
        kwargs = data.get("kwargs", {})
        ret = f"Unsupported task '{task}'"
        if task == "get_inventory":
            name = kwargs["name"]
            try:
                ret = self.inventory.workers[name]
            except KeyError:
                log.error(f"sid.service.broker - no inventory data found for '{name}'")
                ret = None

        reply = orjson.dumps(ret)
        self.send_to_client(
            sender, NFP.RESPONSE, b"sid.service.broker", [uuid, b"200", reply]
        )
        self.send_to_client(
            sender, NFP.MMI, b"sid.service.broker", [uuid, b"200", reply]
        )



import logging

from typing import Union, Any
from norfab.models import Result
from norfab.core.worker import Task, Job
from norfab.core.exceptions import UnsupportedPluginError
from nornir_salt.plugins.tasks import (
    napalm_configure,
    netmiko_send_config,
    scrapli_send_config,
    nr_test,
)
from nornir_salt.plugins.functions import ResultSerializer

log = logging.getLogger(__name__)


class CfgTask:
    @Task(
        fastapi={"methods": ["POST"]},
    )
    def cfg(
        self,
        job: Job,
        config: Union[str, list],
        plugin: str = "netmiko",
        dry_run: bool = False,
        to_dict: bool = True,
        add_details: bool = False,
        job_data: Any = None,
        **kwargs: Any,
    ) -> Result:
        """
        Task to send configuration commands to devices using Command Line Interface (CLI).

        Args:
            job: NorFab Job object containing relevant metadata
            config (list): List of commands to send to devices or URL to a file or template
                URL that resolves to a file.
            plugin (str, optional): Plugin name to use. Valid options are:

                - netmiko - use Netmiko to configure devices
                - scrapli - use Scrapli to configure devices
                - napalm - use NAPALM to configure devices

            dry_run (bool, optional): If True, will not send commands to devices but just return them.
            to_dict (bool, optional): If True, returns results as a dictionary. Defaults to True.
            add_details (bool, optional): If True, adds task execution details to the results.
            job_data (str, optional): URL to YAML file with data or dictionary/list of data to pass on to Jinja2 rendering context.
            **kwargs: Additional arguments to pass to the task plugin.

        Returns:
            dict: A dictionary with the results of the configuration task.

        Raises:
            UnsupportedPluginError: If the specified plugin is not supported.
            FileNotFoundError: If the specified job data file cannot be downloaded.
        """
        config = config if isinstance(config, list) else [config]
        ret = Result(task=f"{self.name}:cfg", result={} if to_dict else [])

        filtered_nornir, no_match_result = self.filter_hosts_and_validate(kwargs, ret)
        if ret.status == "no_match":
            return ret

        # decide on what send commands task plugin to use
        if plugin == "netmiko":
            task_plugin = netmiko_send_config
        elif plugin == "scrapli":
            task_plugin = scrapli_send_config
        elif plugin == "napalm":
            task_plugin = napalm_configure
        else:
            raise UnsupportedPluginError(f"Plugin '{plugin}' not supported")

        job_data = self.load_job_data(job_data)

        nr = self._add_processors(filtered_nornir, kwargs, job)  # add processors

        # render config using Jinja2 on a per-host basis
        for host in nr.inventory.hosts.values():
            rendered = self.jinja2_render_templates(
                templates=config,
                context={
                    "host": host,
                    "norfab": self.client,
                    "job_data": job_data,
                    "netbox": self.add_jinja2_netbox(),
                },
                filters=self.add_jinja2_filters(),
            )
            host.data["__task__"] = {"config": rendered}

        # run task
        log.debug(
            f"{self.name} - sending config commands '{config}', kwargs '{kwargs}', is dry_run - '{dry_run}'"
        )
        if dry_run is True:
            result = nr.run(
                task=nr_test, use_task_data="config", name="dry_run", **kwargs
            )
        else:
            with self.connections_lock:
                result = nr.run(task=task_plugin, **kwargs)

        ret.failed = result.failed  # failed is true if any of the hosts failed
        ret.result = ResultSerializer(result, to_dict=to_dict, add_details=add_details)

        # remove __task__ data
        for host_name, host_object in nr.inventory.hosts.items():
            _ = host_object.data.pop("__task__", None)

        self.watchdog.connections_update(nr, plugin)
        self.watchdog.connections_clean()

        return ret



import json
import builtins

from pydantic import (
    StrictStr,
    Field,
)
from ..common import ClientRunJobArgs, log_error_or_result, listen_events
from .nornir_picle_shell_common import (
    NorniHostsFilters,
    TabulateTableModel,
    NornirCommonArgs,
    print_nornir_results,
)
from nornir_salt.plugins.functions import TabulateFormatter
from picle.models import PipeFunctionsModel, Outputters


class NornirTaskShell(
    NorniHostsFilters, TabulateTableModel, NornirCommonArgs, ClientRunJobArgs
):
    plugin: StrictStr = Field(
        ...,
        description="Nornir task.plugin.name to import or nf://path/to/plugin/file.py",
    )
    arguments: StrictStr = Field(
        None,
        description="Plugin arguments JSON formatted string",
    )

    @staticmethod
    def source_plugin():
        return ClientRunJobArgs.walk_norfab_files()

    @staticmethod
    @listen_events
    def run(uuid, *args, **kwargs):
        NFCLIENT = builtins.NFCLIENT
        workers = kwargs.pop("workers", "all")
        timeout = kwargs.pop("timeout", 600)
        verbose_result = kwargs.pop("verbose_result", False)
        nowait = kwargs.pop("nowait", False)

        # handle task argument
        arguments = json.loads(kwargs.pop("arguments", "{}"))
        if arguments and isinstance(arguments, dict):
            kwargs.update(arguments)

        # extract Tabulate arguments
        table = kwargs.pop("table", {})  # tabulate
        headers = kwargs.pop("headers", "keys")  # tabulate
        headers_exclude = kwargs.pop("headers_exclude", [])  # tabulate
        sortby = kwargs.pop("sortby", "host")  # tabulate
        reverse = kwargs.pop("reverse", False)  # tabulate

        if table:
            kwargs["add_details"] = True
            kwargs["to_dict"] = False

        result = NFCLIENT.run_job(
            "nornir",
            "task",
            workers=workers,
            args=args,
            kwargs=kwargs,
            uuid=uuid,
            timeout=timeout,
            nowait=nowait,
        )

        if nowait:
            return result, Outputters.outputter_nested

        result = log_error_or_result(result, verbose_result=verbose_result)

        # form table results
        if table:
            table_data = []
            for w_name, w_res in result.items():
                for item in w_res:
                    item["worker"] = w_name
                    table_data.append(item)
            ret = TabulateFormatter(
                table_data,
                tabulate=table,
                headers=headers,
                headers_exclude=headers_exclude,
                sortby=sortby,
                reverse=reverse,
            )
        else:
            ret = result

        return ret

    class PicleConfig:
        subshell = True
        prompt = "nf[nornir-task]#"
        outputter = print_nornir_results
        pipe = PipeFunctionsModel



"""
Common Pydantic Models for PICLE Client Shells
"""

import logging
import builtins
import time
from datetime import datetime
import threading
import functools
import json
import queue
from uuid import uuid4  # random uuid
from pydantic import (
    BaseModel,
    StrictBool,
    StrictInt,
    StrictStr,
    Field,
)
from enum import Enum
from typing import Union, Optional, List
from rich.console import Console

log = logging.getLogger(__name__)

# ---------------------------------------------------------------------------------------------
# COMMON FUNCTIONS
# ---------------------------------------------------------------------------------------------


def listen_events_thread(uuid, stop, NFCLIENT):
    """Helper function to pretty print events to command line"""
    richconsole = Console()
    start_time = time.time()
    time_fmt = "%d-%b-%Y %H:%M:%S.%f"
    richconsole.print(
        "-" * 45 + " Job Events " + "-" * 47 + "\n"
        f"{datetime.now().strftime(time_fmt)[:-3]} {uuid} job started"
    )
    while not (stop.is_set() or NFCLIENT.exit_event.is_set()):
        try:
            event = NFCLIENT.event_queue.get(block=True, timeout=0.1)
            NFCLIENT.event_queue.task_done()
        except queue.Empty:
            continue
        (
            empty,
            header,
            command,
            service,
            job_uuid,
            status,
            data,
        ) = event
        if job_uuid != uuid.encode("utf-8"):
            NFCLIENT.event_queue.put(event)
            continue

        # extract event parameters
        data = json.loads(data)
        service = data["service"]
        worker = data["worker"]
        task = data["task"]
        timestamp = data["timestamp"]
        message = data["message"]
        # color severity
        severity = data["severity"]
        severity = severity.replace("DEBUG", "[cyan]DEBUG[/cyan]")
        severity = severity.replace("INFO", "[green]INFO[/green]")
        severity = severity.replace("WARNING", "[yellow]WARNING[/yellow]")
        severity = severity.replace("CRITICAL", "[red]CRITICAL[/red]")
        # color status
        status = data["status"]
        status = status.replace("started", "[cyan]started[/cyan]")
        status = status.replace("completed", "[green]completed[/green]")
        status = status.replace("failed", "[red]failed[/red]")
        resource = data["resource"]
        if isinstance(resource, list):
            resource = ", ".join(resource)
        # log event message
        richconsole.print(
            f"{timestamp} {severity} {worker} {status} {service}.{task} {resource} - {message}"
        )

    elapsed = round(time.time() - start_time, 3)
    richconsole.print(
        f"{datetime.now().strftime(time_fmt)[:-3]} {uuid} job completed in {elapsed} seconds\n\n"
        + "-" * 45
        + " Job Results "
        + "-" * 44
        + "\n"
    )


def listen_events(fun):
    """Decorator to listen for events and print them to console"""

    @functools.wraps(fun)
    def wrapper(*args, **kwargs):
        NFCLIENT = builtins.NFCLIENT
        events_thread_stop = threading.Event()
        uuid = uuid4().hex
        progress = kwargs.get("progress", True)
        nowait = kwargs.get("nowait", False)

        # start events thread to handle job events printing
        if progress and nowait is False:
            events_thread = threading.Thread(
                target=listen_events_thread,
                name="NornirCliShell_events_listen_thread",
                args=(
                    uuid,
                    events_thread_stop,
                    NFCLIENT,
                ),
            )
            events_thread.start()

        # run decorated function
        try:
            res = fun(uuid, *args, **kwargs)
        finally:
            # stop events thread
            if NFCLIENT and progress and nowait is False:
                events_thread_stop.set()
                events_thread.join()

        return res

    return wrapper


def log_error_or_result(
    data: dict, verbose_result: bool = False, verbose_on_fail: bool = False
) -> dict:
    """
    Logs errors or messages from the provided data dictionary and returns a dictionary of results based on verbosity settings.

    Args:
        data (dict): A dictionary where each key is a worker name and each
            value is a dictionary containing job result
        verbose_result (bool, optional): If True, includes the full result
            dictionary for each worker in the return value
        verbose_on_fail (bool, optional): If True, includes the full result
            dictionary for failed tasks

    Returns:
        dict: A dictionary containing either the full result or just the "result"
            field for each worker, depending on verbosity settings.

    Logs:
        - Errors if present in the worker's result.
        - Informational messages if present and no errors exist.
    """
    ret = {}

    if data is None:
        log.error("Result data is empty.")
        return
    if not isinstance(data, dict):
        log.error(f"Data is not a dictionary but '{type(data)}'")
        return data

    for w_name, w_res in data.items():
        # decide what to log
        if w_res["errors"]:
            errors = "\n".join(w_res["errors"])
            log.error(f"{w_name} '{w_res['task']}' errors:\n{errors}")
        elif w_res["messages"]:
            messages = "\n".join(w_res["messages"])
            log.info(f"{w_name} '{w_res['task']}' messages:\n{messages}")

        # decide what results to return
        if verbose_result:
            ret[w_name] = w_res
        elif verbose_on_fail and w_res["failed"] is True:
            ret[w_name] = w_res
        else:
            ret[w_name] = w_res["result"]

    return ret


# ---------------------------------------------------------------------------------------------
# COMMON MODELS
# ---------------------------------------------------------------------------------------------


class BoolEnum(Enum):
    TRUE = True
    FALSE = False


class ClientRunJobArgs(BaseModel):
    timeout: Optional[StrictInt] = Field(None, description="Job timeout")
    workers: Union[StrictStr, List[StrictStr]] = Field(
        "all", description="Filter workers to target"
    )
    verbose_result: StrictBool = Field(
        False,
        description="Control output details",
        json_schema_extra={"presence": True},
        alias="verbose-result",
    )
    progress: Optional[StrictBool] = Field(
        True,
        description="Display progress events",
        json_schema_extra={"presence": True},
    )
    nowait: Optional[StrictBool] = Field(
        False,
        description="Do not wait for job to complete",
        json_schema_extra={"presence": True},
    )

    @staticmethod
    def walk_norfab_files():
        NFCLIENT = builtins.NFCLIENT
        response = NFCLIENT.run_job("filesharing", "walk", kwargs={"url": "nf://"})
        wname, wres = next(iter(response.items()))
        return wres["result"]



import logging
import sys
import threading
import time
import os
import signal
import importlib.metadata
import uvicorn

from norfab.core.worker import NFPWorker, Task, Job
from norfab.models import Result
from typing import Union, List, Dict, Any, Annotated, Optional
from diskcache import FanoutCache
from pydantic import BaseModel
from datetime import datetime
from fastapi import Depends, FastAPI, HTTPException, Body, Request
from fastapi.security.http import HTTPAuthorizationCredentials, HTTPBearer
from fastapi.openapi.utils import get_openapi
from starlette import status
from starlette.routing import Route

SERVICE = "fastapi"
API_TITLE = "NORFAB REST API"

log = logging.getLogger(__name__)


class UnauthorizedMessage(BaseModel):
    detail: str = "Bearer token missing or unknown"


def create_api_endpoint(
    service: str, task_name: str, schema: dict, worker: object
) -> callable:
    """
    Creates an asynchronous FastAPI endpoint function for a given service task.

    Args:
        service (str): Service name.
        task_name (str): Task name.
        schema (dict): Input schema for the task.
        worker (object): Worker instance.

    Returns:
        function: An asynchronous endpoint function

    The generated endpoint expects a JSON body containing arguments for
    the job and returns the result of the job execution.
    """
    # We will handle a missing token ourselves
    get_bearer_token = HTTPBearer(auto_error=False)
    default_workers = schema["properties"].get("workers", {}).get("default", "all")

    def get_token(
        auth: Optional[HTTPAuthorizationCredentials] = Depends(get_bearer_token),
    ) -> str:
        # check token exists in database
        if (
            auth is None
            or worker.bearer_token_check(auth.credentials, Job()).result is False
        ):
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail=UnauthorizedMessage().detail,
            )
        return auth.credentials

    async def endpoint(
        request: Request,
        token: str = Depends(get_token),
    ) -> Dict[Annotated[str, Body(description="Worker Name")], Result]:
        kwargs = await request.json()
        workers = kwargs.pop("workers", default_workers)

        log.debug(
            f"FastAPI running '{service}:{task_name}' task, on '{workers}' workers, job data: '{kwargs}'"
        )

        res = worker.client.run_job(
            service=service,
            task=task_name,
            kwargs=kwargs,
            workers=workers,
        )
        return res

    return endpoint


def make_openapi_schema(
    app, regenerate: bool = False, json_refs: Optional[dict] = None
) -> Dict:
    """
    Generates and returns the OpenAPI schema for a FastAPI application.

    Args:
        app (FastAPI): The FastAPI application instance.
        regenerate (bool, optional): If True, forces regeneration of the OpenAPI schema.
        json_refs (dict, optional): A dictionary of JSON references to include in
            the schema under the "$defs" key.

    Returns:
        dict: The generated OpenAPI schema as a dictionary.
    """
    # make app to re-generate openapi schema
    if regenerate is True:
        app.openapi_schema = None
        app.setup()

    openapi_schema = get_openapi(title=API_TITLE, version="1", routes=app.routes)

    # add json references
    if json_refs:
        openapi_schema["$defs"] = json_refs

    return openapi_schema


def service_tasks_api_discovery(
    worker, cycles: int = 30, discover_service: str = "all"
) -> Dict:
    """
    Periodically discovers available service tasks and dynamically registers
    FastAPI endpoints for them.

    This function performs the following steps in a loop:

    1. Retrieves a list of available services from the worker's client.
    2. For each service, fetches its available tasks.
    3. For each task, checks if it should be exposed via FastAPI (i.e.,
        `task["fastapi"]` is not False).
    4. If the corresponding API endpoint does not already exist, registers a new
        FastAPI route for the task, using its input schema and metadata.
    5. Forces regeneration of the OpenAPI schema after new endpoints are added.

    The loop runs on fastapi service startup up to 30 cycles or until the worker's
    exit event is set, with a 10-second delay between cycles.
    """
    result = {}
    json_refs = {}  # dictionary to store JSON references
    while not worker.exit_event.is_set() and cycles > 0:
        tasks = []
        services = []
        try:
            # get a list of workers and construct a list of services
            services = worker.client.mmi("mmi.service.broker", "show_workers")
            services = [
                s["service"]
                for s in services["results"]
                if discover_service == "all" or s["service"] == discover_service
            ]

            # retrieve NorFab services and their tasks
            for service in services:
                # skip already discovered services
                if service in result:
                    continue
                service_tasks = worker.client.run_job(
                    service=service,
                    task="list_tasks",
                    workers="any",
                    timeout=3,
                )
                # skip if client request timed out
                if service_tasks is None:
                    continue
                for wres in service_tasks.values():
                    for t in wres["result"]:
                        t["service"] = service
                    tasks.extend(wres["result"])

            for task in tasks:
                # skip task endpoint creation if set to false
                if task["fastapi"] is False:
                    continue
                # save service to results
                result.setdefault(task["service"], [])
                # continue with creating API endpoint for task
                path = f"{worker.api_prefix}/{task['service']}/{task['name']}/"
                for route in worker.app.routes:
                    if isinstance(route, Route) and route.path == path:
                        break  # do no re-create existing endpoints
                else:
                    # form OpenAPI schema for API endpoint
                    schema = task["inputSchema"]
                    fastapi_schema = task["fastapi"].pop("schema", {"properties": {}})
                    schema["properties"] = {
                        **fastapi_schema["properties"],
                        **schema["properties"],
                    }
                    _ = schema["properties"].pop("job", None)
                    # extract json references
                    if "$defs" in schema:
                        json_refs.update(schema.pop("$defs"))
                    # form add_api_route arguments
                    task["fastapi"].setdefault("methods", ["POST"])
                    task["fastapi"].setdefault("path", path)
                    task["fastapi"].setdefault("description", task["description"])
                    task["fastapi"].setdefault("name", task["name"])
                    # register API endpoint
                    log.debug(
                        f"Registering API endpoint {task['fastapi']['path']}, schema: {schema}"
                    )
                    worker.app.add_api_route(
                        endpoint=create_api_endpoint(
                            service=task["service"],
                            task_name=task["name"],
                            schema=schema,
                            worker=worker,
                        ),
                        responses={
                            status.HTTP_401_UNAUTHORIZED: dict(
                                model=UnauthorizedMessage
                            )
                        },
                        openapi_extra={
                            "requestBody": {
                                "required": True,
                                "content": {"application/json": {"schema": schema}},
                            }
                        },
                        tags=[f"NORFAB {task['service'].upper()}"],
                        **task["fastapi"],
                    )
                    worker.app.openapi_schema = make_openapi_schema(
                        app=worker.app, regenerate=True, json_refs=json_refs
                    )
                    # save discovered task to results
                    result[task["service"]].append(task["fastapi"]["name"])
        except Exception as e:
            log.exception(f"Failed to discover services tasks, error: {e}")

        cycles -= 1
        time.sleep(10)

    return result


class FastAPIWorker(NFPWorker):
    """
    FastAPIWorker is a worker class that integrates with FastAPI and Uvicorn to serve a FastAPI application.
    It handles initialization, starting the server, and managing bearer tokens.

    Args:
        inventory (str): Inventory configuration for the worker.
        broker (str): Broker URL to connect to.
        worker_name (str): Name of this worker.
        exit_event (threading.Event, optional): Event to signal worker to stop/exit.
        init_done_event (threading.Event, optional): Event to signal when worker is done initializing.
        log_level (str, optional): Logging level for this worker.
        log_queue (object, optional): Queue for logging.
    """

    def __init__(
        self,
        inventory: str,
        broker: str,
        worker_name: str,
        exit_event=None,
        init_done_event=None,
        log_level: str = None,
        log_queue: object = None,
    ):
        super().__init__(
            inventory, broker, SERVICE, worker_name, exit_event, log_level, log_queue
        )
        self.init_done_event = init_done_event
        self.exit_event = exit_event
        self.api_prefix = "/api"

        # get inventory from broker
        self.fastapi_inventory = self.load_inventory()
        self.uvicorn_inventory = {
            "host": "0.0.0.0",
            "port": 8000,
            **self.fastapi_inventory.pop("uvicorn", {}),
        }

        # instantiate cache
        self.cache_dir = os.path.join(self.base_dir, "cache")
        os.makedirs(self.cache_dir, exist_ok=True)
        self.cache = self._get_diskcache()
        self.cache.expire()

        # start FastAPI server
        self.fastapi_start()

        self.service_tasks_api_discovery_thread = threading.Thread(
            target=service_tasks_api_discovery, args=(self,)
        )
        self.service_tasks_api_discovery_thread.start()

        self.init_done_event.set()

    def _get_diskcache(self) -> FanoutCache:
        """
        Initializes and returns a FanoutCache object.

        The FanoutCache is configured with the following parameters:

        - directory: The directory where the cache will be stored.
        - shards: Number of shards to use for the cache.
        - timeout: Timeout for cache operations in seconds.
        - size_limit: Maximum size of the cache in bytes.

        Returns:
            FanoutCache: An instance of FanoutCache configured with the specified parameters.
        """
        return FanoutCache(
            directory=self.cache_dir,
            shards=4,
            timeout=1,  # 1 second
            size_limit=1073741824,  #  1 GigaByte
        )

    def fastapi_start(self):
        """
        Starts the FastAPI server.

        This method initializes the FastAPI application using the provided
        configuration, starts the Uvicorn server in a separate thread, and waits
        for the server to be fully started before logging the server's URL.

        Steps:

        1. Create the FastAPI application using `make_fast_api_app`.
        2. Configure the Uvicorn server with the application and settings.
        3. Start the Uvicorn server in a new thread.
        4. Wait for the server to start.
        5. Log the server's URL.'

        Attributes:
            self.app (FastAPI): The FastAPI application instance.
            self.uvicorn_server (uvicorn.Server): The Uvicorn server instance.
            self.uvicorn_server_thread (threading.Thread): The thread running the Uvicorn server.

        Raises:
            Exception: If the server fails to start.
        """
        self.app = make_fast_api_app(
            worker=self, config=self.fastapi_inventory.get("fastapi", {})
        )

        # start uvicorn server in a thread
        config = uvicorn.Config(app=self.app, **self.uvicorn_inventory)
        self.uvicorn_server = uvicorn.Server(config=config)

        self.uvicorn_server_thread = threading.Thread(target=self.uvicorn_server.run)
        self.uvicorn_server_thread.start()

        # wait for server to start
        while not self.uvicorn_server.started:
            time.sleep(0.001)

        log.info(
            f"{self.name} - Uvicorn server started, serving FastAPI app at "
            f"http://{self.uvicorn_inventory['host']}:{self.uvicorn_inventory['port']}"
        )

    def worker_exit(self):
        """
        Terminates the current process by sending a SIGTERM signal to itself.

        This method retrieves the current process ID using `os.getpid()` and then
        sends a SIGTERM signal to terminate the process using `os.kill()`.
        """
        os.kill(os.getpid(), signal.SIGTERM)

    @Task(fastapi={"methods": ["GET"]})
    def get_version(self) -> Result:
        """
        Produce a report of the versions of various Python packages.

        This method collects the versions of several specified Python packages
        and returns them in a dictionary.

        Returns:
            Result: An object containing the task name and a dictionary with
                    the package names as keys and their respective versions as values.
        """
        libs = {
            "norfab": "",
            "fastapi": "",
            "uvicorn": "",
            "pydantic": "",
            "python-multipart": "",
            "python": sys.version.split(" ")[0],
            "platform": sys.platform,
        }
        # get version of packages installed
        for pkg in libs.keys():
            try:
                libs[pkg] = importlib.metadata.version(pkg)
            except importlib.metadata.PackageNotFoundError:
                pass

        return Result(task=f"{self.name}:get_version", result=libs)

    @Task(fastapi={"methods": ["GET"]})
    def get_inventory(self) -> Result:
        """
        Retrieve the inventory of the FastAPI worker.

        Returns:
            Dict: A dictionary containing the combined inventory of FastAPI and Uvicorn.
        """
        return Result(
            result={**self.fastapi_inventory, "uvicorn": self.uvicorn_inventory},
            task=f"{self.name}:get_inventory",
        )

    @Task(fastapi={"methods": ["GET"]})
    def get_openapi_schema(self, paths: bool = False) -> Result:
        """
        Generates and returns the OpenAPI schema for the FastAPI application.

        Args:
            paths (bool, optional): If True, returns a list of available API endpoint paths.
                If False, returns the full OpenAPI schema. Defaults to False.

        Returns:
            Result: An object containing either the list of endpoint paths or the full OpenAPI schema
        """
        schema = make_openapi_schema(self.app)
        if paths is True:
            return Result(
                result=list(schema["paths"].keys()),
                task=f"{self.name}:get_openapi_schema",
            )
        else:
            return Result(
                result=schema,
                task=f"{self.name}:get_openapi_schema",
            )

    @Task(fastapi=False, mcp=False)
    def bearer_token_store(
        self, job: Job, username: str, token: str, expire: int = None
    ) -> Result:
        """
        Method to store a bearer token in the database.

        This method stores a bearer token associated with a username in the cache.

        If an expiration time is not provided, it retrieves the default token TTL
        from the FastAPI inventory configuration.

        Args:
            username: str - The name of the user to store the token for.
            token: str - The token string to store.
            expire: int, optional - The number of seconds before the token expires.

        Returns:
            bool - Returns True if the token is successfully stored.
        """
        expire = expire or self.fastapi_inventory.get("auth_bearer", {}).get(
            "token_ttl", expire
        )
        self.cache.expire()
        cache_key = f"bearer_token::{token}"
        if cache_key in self.cache:
            user_token = self.cache.get(cache_key)
        else:
            user_token = {
                "token": token,
                "username": username,
                "created": str(datetime.now()),
            }
        self.cache.set(cache_key, user_token, expire=expire, tag=username)

        return Result(task=f"{self.name}:bearer_token_store", result=True)

    @Task(fastapi=False, mcp=False)
    def bearer_token_delete(
        self, job: Job, username: str = None, token: str = None
    ) -> Result:
        """
        Deletes a bearer token from the cache.
        This method removes a bearer token from the cache based on either
        the token itself or the associated username.

        If a token is provided, it will be removed directly. If a username
        is provided, all tokens associated with that username will be evicted
        from the cache.

        Args:
            username (str, optional): The username associated with the token(s) to be removed. Defaults to None.
            token (str, optional): The bearer token to be removed. Defaults to None.

        Returns:
            bool: True if the operation was successful, otherwise raises an exception.

        Raises:
            RuntimeError: If the token removal from the cache fails.
            Exception: If neither username nor token is provided.
        """
        self.cache.expire()
        token_removed_count = 0
        if token:
            cache_key = f"bearer_token::{token}"
            if cache_key in self.cache:
                if self.cache.delete(cache_key, retry=True):
                    token_removed_count = 1
                else:
                    raise RuntimeError(f"Failed to remove {username} token from cache")
        elif username:
            token_removed_count = self.cache.evict(tag=username, retry=True)
        else:
            raise Exception("Cannot delete, either username or token must be provided")

        log.info(
            f"{self.name} removed {token_removed_count} token(s) for user {username}"
        )

        return Result(task=f"{self.name}:bearer_token_delete", result=True)

    @Task(fastapi=False, mcp=False)
    def bearer_token_list(self, job: Job, username: str = None) -> Result:
        """
        Retrieves a list of bearer tokens from the cache, optionally filtered by username.

        Args:
            username (str, optional): The username to filter tokens by. Defaults to None.

        Returns:
            list: A list of dictionaries containing token information. Each dictionary contains:

                - "username" (str): The username associated with the token.
                - "token" (str): The bearer token.
                - "age" (str): The age of the token.
                - "creation" (str): The creation time of the token.
                - "expires" (str): The expiration time of the token, if available.

        If no tokens are found, a list with a single dictionary containing
        empty strings for all fields is returned.
        """

        self.cache.expire()
        ret = Result(task=f"{self.name}:bearer_token_list", result=[])

        for cache_key in self.cache:
            token_data, expires, tag = self.cache.get(
                cache_key, expire_time=True, tag=True
            )
            if username and tag != username:
                continue
            if expires is not None:
                expires = datetime.fromtimestamp(expires)
            creation = datetime.fromisoformat(token_data["created"])
            age = datetime.now() - creation
            ret.result.append(
                {
                    "username": token_data["username"],
                    "token": token_data["token"],
                    "age": str(age),
                    "creation": str(creation),
                    "expires": str(expires),
                }
            )

        # return empty result if no tokens found
        if not ret.result:
            ret.result = [
                {
                    "username": "",
                    "token": "",
                    "age": "",
                    "creation": "",
                    "expires": "",
                }
            ]

        return ret

    @Task(fastapi=False, mcp=False)
    def bearer_token_check(self, token: str, job: Job) -> Result:
        """
        Checks if the provided bearer token is present in the cache and still active.

        Args:
            token (str): The bearer token to check.

        Returns:
            bool: True if the token is found in the cache, False otherwise.
        """
        self.cache.expire()
        cache_key = f"bearer_token::{token}"
        return Result(
            task=f"{self.name}:bearer_token_check", result=cache_key in self.cache
        )

    @Task(fastapi={"methods": ["POST"]})
    def discover(self, job, service: str = "all", progress: bool = True) -> Result:
        """
        Discovers available services tasks and auto-generate API endpoints for them.

        Args:
            service (str, optional): The name of the service to discover. Defaults to "all".

        Returns:
            Result: An object containing the discovery results for the specified service.
        """
        job.event("Discovering NorFab services tasks")
        ret = Result(task=f"{self.name}:discover")
        ret.result = service_tasks_api_discovery(
            self, cycles=1, discover_service=service
        )

        return ret


# ------------------------------------------------------------------
# FastAPI REST API routes endpoints
# ------------------------------------------------------------------


def make_fast_api_app(worker: object, config: dict) -> FastAPI:
    """
    Create a FastAPI application with endpoints for posting, getting, and running jobs.

    This function sets up a FastAPI application with three endpoints:

    - POST /job: To post a job to the NorFab service.
    - GET /job: To get job results from the NorFab service.
    - POST /job/run: To run a job and return job results synchronously.

    Each endpoint requires a bearer token for authentication, which is validated
    against the worker's token database.

    Args:
        worker (object): An object representing the worker that will handle the job requests.
        config (dict): A dictionary of configuration options for the FastAPI application.

    Returns:
        FastAPI: A FastAPI application instance.
    """
    config = {
        "title": API_TITLE,
        "summary": "NorFab Services Tasks FastAPI application with endpoints for posting, getting, and running jobs",
        **config,
    }
    app = FastAPI(**config)

    # We will handle a missing token ourselves
    get_bearer_token = HTTPBearer(auto_error=False)

    def get_token(
        auth: Optional[HTTPAuthorizationCredentials] = Depends(get_bearer_token),
    ) -> str:
        # check token exists in database
        if (
            auth is None
            or worker.bearer_token_check(auth.credentials, Job()).result is False
        ):
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail=UnauthorizedMessage().detail,
            )
        return auth.credentials

    # @app.post(
    #     f"{worker.api_prefix}/job",
    #     responses={status.HTTP_401_UNAUTHORIZED: dict(model=UnauthorizedMessage)},
    #     tags=["NORFAB"],
    # )
    # def post_job(
    #     service: Annotated[
    #         str, Body(description="The name of the service to post the job to")
    #     ],
    #     task: Annotated[
    #         str, Body(description="The task to be executed by the service")
    #     ],
    #     args: Annotated[
    #         List[Any], Body(description="A list of positional arguments for the task")
    #     ] = None,
    #     kwargs: Annotated[
    #         Dict[str, Any],
    #         Body(description="A dictionary of keyword arguments for the task"),
    #     ] = None,
    #     workers: Annotated[
    #         Union[str, List[str]], Body(description="The workers to dispatch the task")
    #     ] = "all",
    #     uuid: Annotated[
    #         str, Body(description="Optional a unique identifier to use for the job")
    #     ] = None,
    #     timeout: Annotated[
    #         int, Body(description="The timeout for the job in seconds")
    #     ] = 600,
    #     token: str = Depends(get_token),
    # ) -> ClientPostJobResponse:
    #     """
    #     Method to post the job to NorFab.
    #
    #     Args:
    #         service: The name of the service to post the job to.
    #         task: The task to be executed by the service.
    #         args: A list of positional arguments for the task. Defaults to None.
    #         kwargs: A dictionary of keyword arguments for the task. Defaults to None.
    #         workers: The workers to dispatch the task. Defaults to "all".
    #         uuid: Optional a unique identifier to use for the job. Defaults to None.
    #         timeout: The timeout for the job in seconds. Defaults to 600.
    #
    #     Returns:
    #         The response from the NorFab service.
    #     """
    #     log.debug(
    #         f"{worker.name} - received job post request, service {service}, task {task}, args {args}, kwargs {kwargs}"
    #     )
    #     res = worker.client.post(
    #         service=service,
    #         task=task,
    #         args=args,
    #         kwargs=kwargs,
    #         workers=workers,
    #         timeout=timeout,
    #         uuid=uuid,
    #     )
    #     return res

    # @app.get(
    #     f"{worker.api_prefix}/job",
    #     responses={status.HTTP_401_UNAUTHORIZED: dict(model=UnauthorizedMessage)},
    #     tags=["NORFAB"],
    # )
    # def get_job(
    #     service: Annotated[
    #         str, Body(description="The name of the service to get the job from")
    #     ],
    #     uuid: Annotated[str, Body(description="A unique identifier for the job")],
    #     workers: Annotated[
    #         Union[str, List[str]],
    #         Body(description="The workers to dispatch the get request to"),
    #     ] = "all",
    #     timeout: Annotated[
    #         int, Body(description="The timeout for the job in seconds")
    #     ] = 600,
    #     token: str = Depends(get_token),
    # ) -> ClientGetJobResponse:
    #     """
    #     Method to get job results from NorFab.
    #
    #     Args:
    #         service: The name of the service to get the job from.
    #         workers: The workers to dispatch the get request to. Defaults to "all".
    #         uuid: A unique identifier for the job.
    #         timeout: The timeout for the job get requests in seconds. Defaults to 600.
    #
    #     Returns:
    #         The response from the NorFab service.
    #     """
    #     log.debug(
    #         f"{worker.name} - received job get request, service {service}, uuid {uuid}"
    #     )
    #     res = worker.client.mmi(
    #         service=service,
    #         uuid=uuid,
    #         workers=workers,
    #         timeout=timeout,
    #     )
    #     return res

    @app.post(
        f"{worker.api_prefix}/job/run",
        responses={status.HTTP_401_UNAUTHORIZED: dict(model=UnauthorizedMessage)},
        tags=["NORFAB"],
    )
    def run_job(
        service: Annotated[
            str, Body(description="The name of the service to post the job to")
        ],
        task: Annotated[
            str, Body(description="The task to be executed by the service")
        ],
        args: Annotated[
            List[Any], Body(description="A list of positional arguments for the task")
        ] = None,
        kwargs: Annotated[
            Dict[str, Any],
            Body(description="A dictionary of keyword arguments for the task"),
        ] = None,
        workers: Annotated[
            Union[str, List[str]], Body(description="The workers to dispatch the task")
        ] = "all",
        uuid: Annotated[
            str, Body(description="Optional a unique identifier to use for the job")
        ] = None,
        timeout: Annotated[
            int, Body(description="The timeout for the job in seconds")
        ] = 600,
        token: str = Depends(get_token),
    ) -> Dict[str, Result]:
        """
        Method to run job and return job results synchronously. This function
        is blocking, internally it uses post/get methods to submit job request
        and waits for job results to come through for all workers request was
        dispatched to, exiting either once timeout expires or after all workers
        reported job result back to the client.

        Args:
            service: The name of the service to post the job to.
            task: The task to be executed by the service.
            args: A list of positional arguments for the task. Defaults to None.
            kwargs: A dictionary of keyword arguments for the task. Defaults to None.
            workers: The workers to dispatch the task. Defaults to "all".
            uuid: A unique identifier for the job. Defaults to None.
            timeout: The timeout for the job in seconds. Defaults to 600.

        Returns:
            The response from the NorFab service.
        """
        log.debug(
            f"{worker.name} - received run job request, service {service}, task {task}, args {args}, kwargs {kwargs}"
        )
        res = worker.client.run_job(
            service=service,
            task=task,
            uuid=uuid,
            args=args,
            kwargs=kwargs,
            workers=workers,
            timeout=timeout,
        )
        return res

    return app



import logging

from typing import Union
from norfab.core.worker import Task, Job
from norfab.models import Result
from .netbox_models import NetboxFastApiArgs
from .netbox_exceptions import UnsupportedNetboxVersion

log = logging.getLogger(__name__)


class NetboxConnectionsTasks:

    @Task(fastapi={"methods": ["GET"], "schema": NetboxFastApiArgs.model_json_schema()})
    def get_connections(
        self,
        job: Job,
        devices: list[str],
        instance: Union[None, str] = None,
        dry_run: bool = False,
        cables: bool = False,
        cache: Union[None, bool, str] = None,
        include_virtual: bool = True,
        interface_regex: Union[None, str] = None,
    ) -> Result:
        """
        Retrieve interface connection details for specified devices from Netbox.

        This task retrieves these connections:

        - Physical interfaces connections
        - Child/virtual interfaces connections using parent interface connections details
        - Lag interfaces connections using member ports connections details
        - Lag child interfaces connections using member ports connections details
        - Console port and console server ports connections
        - Connections to provider networks for physical, child/virtual and lag interfaces

        Args:
            job: NorFab Job object containing relevant metadata
            devices (list): List of device names to retrieve connections for.
            instance (str, optional): Netbox instance name for the GraphQL query.
            dry_run (bool, optional): If True, perform a dry run without making actual changes.
            cables (bool, optional): if True includes interfaces' directly attached cables and peers details
            include_virtual (bool, optional): if True include connections for virtual and LAG interfaces
            interface_regex (str, optional): Regex pattern to match interfaces, console ports and
                console server ports by name, case insensitive.

        Returns:
            dict: A dictionary containing connection details for each device:

                ```
                {
                    "netbox-worker-1.2": {
                        "r1": {
                            "Console": {
                                "breakout": false,
                                "remote_device": "termserv1",
                                "remote_device_status": "active",
                                "remote_interface": "ConsoleServerPort1",
                                "remote_termination_type": "consoleserverport",
                                "termination_type": "consoleport"
                            },
                            "eth1": {
                                "breakout": false,
                                "remote_device": "r2",
                                "remote_device_status": "active",
                                "remote_interface": "eth8",
                                "remote_termination_type": "interface",
                                "termination_type": "interface"
                            }
                        }
                    }
                }
                ```

        Raises:
            Exception: If there is an error in the GraphQL query or data retrieval process.
        """
        instance = instance or self.default_instance
        ret = Result(
            task=f"{self.name}:get_connections",
            result={d: {} for d in devices},
            resources=[instance],
        )

        # form lists of fields to request from netbox
        cable_fields = """
            cable {
                type
                status
                tenant {name}
                label
                tags {name}
                length
                length_unit
                custom_fields
            }
        """
        interfaces_fields = [
            "name",
            "type",
            "device {name, status}",
            """
            member_interfaces {
              name
              connected_endpoints {
                __typename
                ... on ProviderNetworkType {name}
                ... on InterfaceType {name, device {name, status}, child_interfaces {name}, lag {name child_interfaces {name}}}
              }
            }
            """,
            """
            parent {
              name
              type
              member_interfaces {
                name
                connected_endpoints {
                  __typename
                  ... on ProviderNetworkType {name}
                  ... on InterfaceType {name, device {name, status}, child_interfaces {name}, lag {name child_interfaces {name}}}
                }
              }
              connected_endpoints {
                __typename
                ... on ProviderNetworkType {name}
                ... on InterfaceType {name, device {name, status}, child_interfaces {name}, lag {name child_interfaces {name}}}
              }
            }
            """,
            """
            connected_endpoints {
                __typename 
                ... on ProviderNetworkType {name}
                ... on InterfaceType {name, device {name, status}, child_interfaces {name}, lag {name child_interfaces {name}}}
            }
            """,
            """
            link_peers {
                __typename
                ... on InterfaceType {name device {name, status}}
                ... on FrontPortType {name device {name, status}}
                ... on RearPortType {name device {name, status}}
            }
            """,
        ]
        console_ports_fields = [
            "name",
            "device {name, status}",
            "type",
            """connected_endpoints {
              __typename 
              ... on ConsoleServerPortType {name device {name, status}}
            }""",
            """link_peers {
              __typename
              ... on ConsoleServerPortType {name device {name, status}}
              ... on FrontPortType {name device {name, status}}
              ... on RearPortType {name device {name, status}}
            }""",
        ]
        console_server_ports_fields = [
            "name",
            "device {name, status}",
            "type",
            """connected_endpoints {
              __typename 
              ... on ConsolePortType {name device {name, status}}
            }""",
            """link_peers {
              __typename
              ... on ConsolePortType {name device {name, status}}
              ... on FrontPortType {name device {name, status}}
              ... on RearPortType {name device {name, status}}
            }""",
        ]
        power_outlet_fields = [
            "name",
            "device {name, status}",
            "type",
            """connected_endpoints {
              __typename 
              ... on PowerPortType {name device {name, status}}
            }""",
            """link_peers {
              __typename
              ... on PowerPortType {name device {name, status}}
            }""",
        ]

        # check if need to include cables info
        if cables is True:
            interfaces_fields.append(cable_fields)
            console_ports_fields.append(cable_fields)
            console_server_ports_fields.append(cable_fields)
            power_outlet_fields.append(cable_fields)

        # form query dictionary with aliases to get data from Netbox
        dlist = str(devices).replace("'", '"')  # swap quotes
        if self.nb_version[instance] >= (4, 4, 0):
            if interface_regex:
                filters = (
                    "{device: {name: {in_list: "
                    + dlist
                    + "}}, "
                    + "name: {i_regex: "
                    + f'"{interface_regex}"'
                    + "}}"
                )
            else:
                filters = "{device: {name: {in_list: " + dlist + "}}}"
        else:
            raise UnsupportedNetboxVersion(
                f"{self.name} - Netbox version {self.nb_version[instance]} is not supported, "
                f"minimum required version is {self.compatible_ge_v4}"
            )

        queries = {
            "interface": {
                "obj": "interface_list",
                "filters": filters,
                "fields": interfaces_fields,
            },
            "consoleport": {
                "obj": "console_port_list",
                "filters": filters,
                "fields": console_ports_fields,
            },
            "consoleserverport": {
                "obj": "console_server_port_list",
                "filters": filters,
                "fields": console_server_ports_fields,
            },
            "poweroutlet": {
                "obj": "power_outlet_list",
                "filters": filters,
                "fields": power_outlet_fields,
            },
        }

        # retrieve full list of devices interface with all cables
        query_result = self.graphql(
            job=job, queries=queries, instance=instance, dry_run=dry_run
        )

        # return dry run result
        if dry_run:
            return query_result

        all_ports = query_result.result
        if not all_ports:
            return ret

        # extract physical interfaces connections
        for port_type, ports in all_ports.items():
            for port in ports:
                # skip ports that have no remote device connected
                endpoints = port["connected_endpoints"]
                if not endpoints or not all(i for i in endpoints):
                    continue

                # extract required parameters
                cable = port.get("cable", {})
                device_name = port["device"]["name"]
                port_name = port["name"]
                link_peers = port["link_peers"]
                remote_termination_type = endpoints[0]["__typename"].lower()
                remote_termination_type = remote_termination_type.replace("type", "")

                # form initial connection dictionary
                connection = {
                    "breakout": len(endpoints) > 1,
                    "remote_termination_type": remote_termination_type,
                    "termination_type": port_type,
                }

                # add remote connection details
                if remote_termination_type == "providernetwork":
                    connection["remote_device"] = None
                    connection["remote_device_status"] = None
                    connection["remote_interface"] = None
                    connection["provider"] = endpoints[0]["name"]
                else:
                    remote_interface = endpoints[0]["name"]
                    if len(endpoints) > 1:
                        remote_interface = list(sorted([i["name"] for i in endpoints]))
                    connection["remote_interface"] = remote_interface
                    connection["remote_device"] = endpoints[0]["device"]["name"]
                    connection["remote_device_status"] = endpoints[0]["device"][
                        "status"
                    ]

                # add cable and its peer details
                if cables:
                    peer_termination_type = link_peers[0]["__typename"].lower()
                    peer_termination_type = peer_termination_type.replace("type", "")
                    cable["peer_termination_type"] = peer_termination_type
                    cable["peer_device"] = link_peers[0].get("device", {}).get("name")
                    cable["peer_interface"] = link_peers[0].get("name")
                    if len(link_peers) > 1:  # handle breakout cable
                        cable["peer_interface"] = [i["name"] for i in link_peers]
                    connection["cable"] = cable

                # add physical connection to the results
                ret.result[device_name][port_name] = connection

        # extract virtual interfaces connections
        for port_type, ports in all_ports.items():
            for port in ports:
                # add child virtual interfaces connections
                if (
                    not include_virtual
                    or port["type"] != "virtual"
                    or not port["parent"]
                ):
                    continue
                device_name = port["device"]["name"]
                interface_name = port["name"]
                parent = port["parent"]
                connection = {
                    "remote_device": None,
                    "remote_device_status": None,
                    "remote_interface": None,
                    "remote_termination_type": "virtual",
                    "termination_type": "virtual",
                }
                # find connection endpoint
                if parent["type"] == "lag":
                    try:
                        endpoint = parent["member_interfaces"][0][
                            "connected_endpoints"
                        ][0]
                    except Exception:
                        continue
                elif parent["connected_endpoints"]:
                    try:
                        endpoint = parent["connected_endpoints"][0]
                    except Exception:
                        continue
                else:
                    log.error(
                        f"{device_name}:{interface_name} parent has no connected endpoints"
                    )
                    continue
                connection["remote_device"] = endpoint["device"]["name"]
                connection["remote_device_status"] = endpoint["device"]["status"]
                remote_termination_type = endpoint["__typename"].lower()
                remote_termination_type = remote_termination_type.replace("type", "")
                # collect virtual interfaces facing provider
                if remote_termination_type == "providernetwork":
                    connection["provider"] = endpoint["name"]
                # find matching remote virtual interface for LAG subif
                elif "." in interface_name and parent["type"] == "lag":
                    subif_id = interface_name.split(".")[1]
                    for remote_child in endpoint["lag"]["child_interfaces"]:
                        if remote_child["name"].endswith(f".{subif_id}"):
                            connection["remote_interface"] = remote_child["name"]
                            break
                    # no matching subinterface found, associate child interface with remote interface
                    else:
                        connection["remote_interface"] = endpoint["lag"]["name"]
                        connection["remote_termination_type"] = "lag"
                # find matching remote virtual interface for physical interface subif
                elif "." in interface_name:
                    subif_id = interface_name.split(".")[1]
                    for remote_child in endpoint["child_interfaces"]:
                        if remote_child["name"].endswith(f".{subif_id}"):
                            connection["remote_interface"] = remote_child["name"]
                            break
                    # no matching subinterface found, associate child interface with remote interface
                    else:
                        connection["remote_interface"] = endpoint["name"]
                        connection["remote_termination_type"] = remote_termination_type
                # add virtual interface connection to results
                ret.result[device_name][interface_name] = connection

        # extract LAG interfaces connections
        for port_type, ports in all_ports.items():
            for port in ports:
                if not include_virtual or port["type"] != "lag":
                    continue
                device_name = port["device"]["name"]
                interface_name = port["name"]
                connection = {
                    "remote_device": None,
                    "remote_device_status": None,
                    "remote_interface": None,
                    "remote_termination_type": "lag",
                    "termination_type": "lag",
                }
                try:
                    endpoint = port["member_interfaces"][0]["connected_endpoints"][0]
                except Exception:
                    continue
                remote_termination_type = endpoint["__typename"].lower()
                remote_termination_type = remote_termination_type.replace("type", "")
                # collect lag interfaces facing provider
                if remote_termination_type == "providernetwork":
                    connection["provider"] = endpoint["name"]
                # find remote lag interface
                elif endpoint["lag"]:
                    connection["remote_interface"] = endpoint["lag"]["name"]
                    connection["remote_device"] = endpoint["device"]["name"]
                    connection["remote_device_status"] = endpoint["device"]["status"]
                # add lag interface connection to results
                ret.result[device_name][interface_name] = connection

        return ret



from norfab.models import Result
from norfab.core.worker import Task, Job
from nornir_salt.plugins.functions import InventoryFun


class RuntimeInventoryTask:
    @Task(fastapi={"methods": ["POST"]})
    def runtime_inventory(self, job: Job, action: str, **kwargs) -> Result:
        """
        Task to work with Nornir runtime (in-memory) inventory.

        Supported actions:

        - `create_host` or `create` - creates new host or replaces existing host object
        - `read_host` or `read` - read host inventory content
        - `update_host` or `update` - non recursively update host attributes if host exists
            in Nornir inventory, do not create host if it does not exist
        - `delete_host` or `delete` - deletes host object from Nornir Inventory
        - `load` - to simplify calling multiple functions
        - `read_inventory` - read inventory content for groups, default and hosts
        - `read_host_data` - to return host's data under provided path keys
        - `list_hosts` - return a list of inventory's host names
        - `list_hosts_platforms` - return a dictionary of hosts' platforms
        - `update_defaults` - non recursively update defaults attributes

        Args:
            job: NorFab Job object containing relevant metadata
            action: action to perform on inventory
            kwargs: arguments to use with the calling action
        """
        # clean up kwargs
        _ = kwargs.pop("progress", None)
        job.event(f"Performing '{action}' action")
        return Result(result=InventoryFun(self.nr, call=action, **kwargs))



import builtins

from enum import Enum
from pydantic import (
    BaseModel,
    StrictBool,
    StrictFloat,
    StrictStr,
    Field,
)
from ..common import ClientRunJobArgs, log_error_or_result, listen_events
from .nornir_picle_shell_common import (
    NorniHostsFilters,
    TabulateTableModel,
    NornirCommonArgs,
    print_nornir_results,
)
from nornir_salt.plugins.functions import TabulateFormatter
from picle.models import PipeFunctionsModel, Outputters


class SCPDirection(str, Enum):
    put = "put"
    get = "get"


class NrFileCopyPluginNetmiko(BaseModel):
    dest_file: StrictStr = Field(
        None, description="Destination file to copy", alias="destination-file"
    )
    file_system: StrictStr = Field(
        None, description="Destination file system", alias="file-system"
    )
    direction: SCPDirection = Field("put", description="Direction of file copy")
    inline_transfer: StrictBool = Field(
        False,
        description="Use inline transfer, supported by Cisco IOS",
        alias="inline-transfer",
        json_schema_extra={"presence": True},
    )
    overwrite_file: StrictBool = Field(
        False,
        description="Overwrite destination file if it exists",
        alias="overwrite-file",
        json_schema_extra={"presence": True},
    )
    socket_timeout: StrictFloat = Field(
        10.0, description="Socket timeout in seconds", alias="socket-timeout"
    )
    verify_file: StrictBool = Field(
        True,
        description="Verify destination file hash after copy",
        alias="verify-file",
        json_schema_extra={"presence": True},
    )

    @staticmethod
    def run(*args, **kwargs):
        kwargs["plugin"] = "netmiko"
        return NornirFileCopyShell.run(*args, **kwargs)

    class PicleConfig:
        outputter = print_nornir_results


class NrFileCopyPlugins(BaseModel):
    netmiko: NrFileCopyPluginNetmiko = Field(
        None, description="Use Netmiko plugin to copy files"
    )


class NornirFileCopyShell(
    NorniHostsFilters, TabulateTableModel, NornirCommonArgs, ClientRunJobArgs
):
    source_file: StrictStr = Field(
        ..., description="Source file to copy", alias="source-file"
    )
    plugin: NrFileCopyPlugins = Field(None, description="Connection plugin parameters")
    dry_run: StrictBool = Field(
        False,
        description="Do not copy files, just show what would be done",
        alias="dry-run",
        json_schema_extra={"presence": True},
    )

    @staticmethod
    def source_source_file():
        return ClientRunJobArgs.walk_norfab_files()

    @staticmethod
    @listen_events
    def run(uuid, *args, **kwargs):
        NFCLIENT = builtins.NFCLIENT
        workers = kwargs.pop("workers", "all")
        timeout = kwargs.pop("timeout", 600)
        verbose_result = kwargs.pop("verbose_result", False)
        nowait = kwargs.pop("nowait", False)

        # extract Tabulate arguments
        table = kwargs.pop("table", {})  # tabulate
        headers = kwargs.pop("headers", "keys")  # tabulate
        headers_exclude = kwargs.pop("headers_exclude", [])  # tabulate
        sortby = kwargs.pop("sortby", "host")  # tabulate
        reverse = kwargs.pop("reverse", False)  # tabulate

        if table:
            kwargs["add_details"] = True
            kwargs["to_dict"] = False

        result = NFCLIENT.run_job(
            "nornir",
            "file_copy",
            workers=workers,
            args=args,
            kwargs=kwargs,
            uuid=uuid,
            timeout=timeout,
            nowait=nowait,
        )

        if nowait:
            return result, Outputters.outputter_nested

        result = log_error_or_result(result, verbose_result=verbose_result)

        # form table results
        if table:
            table_data = []
            for w_name, w_res in result.items():
                for item in w_res:
                    item["worker"] = w_name
                    table_data.append(item)
            ret = TabulateFormatter(
                table_data,
                tabulate=table,
                headers=headers,
                headers_exclude=headers_exclude,
                sortby=sortby,
                reverse=reverse,
            )
        else:
            ret = result

        return ret

    class PicleConfig:
        subshell = True
        prompt = "nf[nornir-file-copy]#"
        outputter = print_nornir_results
        pipe = PipeFunctionsModel



import logging
import sys
import threading
import time
import os
import signal
import importlib.metadata

from fnmatch import fnmatch
from norfab.core.worker import NFPWorker, Task
from norfab.models import Result
from diskcache import FanoutCache
from mcp.server.fastmcp import FastMCP
from mcp import types
from typing import Any, Optional

SERVICE = "fastmcp"

log = logging.getLogger(__name__)


def service_tasks_discovery(
    worker: Any, cycles: int = 5, discover_service: str = "all"
) -> dict:
    """
    Discovers available tasks from NorFab services and registers them
    as tools for the worker. This function periodically queries the
    broker for available services and their tasks, and registers each
    discovered task as a tool in the worker's `norfab_services_tasks`
    dictionary. It continues this process for a specified number of
    cycles or until the worker's exit event is set.

    Args:
        worker: The worker instance responsible for managing service
            tasks and tools.
        cycles (int, optional): The number of discovery cycles to perform.
        discover_service (str, optional): The name of a specific service
            to discover tasks from. If set to "all", tasks from all services
            are discovered. Defaults to "all".
    """
    result = {}
    while not worker.exit_event.is_set() and cycles > 0:
        tasks = []
        services = []
        try:
            # get a list of workers and construct a list of services
            services = worker.client.mmi("mmi.service.broker", "show_workers")
            services = [
                s["service"]
                for s in services["results"]
                if discover_service == "all" or s["service"] == discover_service
            ]

            # retrieve NorFab services and their tasks
            for service in services:
                # skip already discovered services
                if service in result:
                    continue
                service_tasks = worker.client.run_job(
                    service=service,
                    task="list_tasks",
                    workers="any",
                    timeout=3,
                )
                # skip if client request timed out
                if service_tasks is None:
                    continue
                for wres in service_tasks.values():
                    for t in wres["result"]:
                        t["service"] = service
                    tasks.extend(wres["result"])

            # create tools for discovered tasks
            for task in tasks:
                # skip task tool creation if set to false
                if task["mcp"] is False:
                    continue
                # save service to results
                result.setdefault(task["service"], {})
                # continue with creating tool for task
                task_tool = {
                    "name": task["name"],
                    "description": task["description"],
                    "inputSchema": task["inputSchema"],
                    "outputSchema": task["outputSchema"],
                    **task["mcp"],
                }
                task_tool["name"] = (
                    f"service_{task['service']}__task_{task_tool['name']}"
                )
                # skip already discovered tasks
                if task_tool["name"] in result[task["service"]]:
                    continue
                # save discovered task to return results
                result[task["service"]][task_tool["name"]] = {
                    "tool": types.Tool(**task_tool),
                    "task": task,
                }
            # save tools to worker tasks dictionary
            worker.norfab_services_tasks.update(result)
        except Exception as e:
            log.exception(f"Failed to discover services tasks, error: {e}")

        cycles -= 1
        time.sleep(5)

    return result


class FastMCPWorker(NFPWorker):

    def __init__(
        self,
        inventory: str,
        broker: str,
        worker_name: str,
        exit_event: Optional[threading.Event] = None,
        init_done_event: Optional[threading.Event] = None,
        log_level: Optional[str] = None,
        log_queue: Optional[object] = None,
    ) -> None:
        super().__init__(
            inventory, broker, SERVICE, worker_name, exit_event, log_level, log_queue
        )
        self.init_done_event = init_done_event
        self.exit_event = exit_event
        self.norfab_services_tasks = {}
        self.mcp_server_name = "NorFab MCP Server"

        # get inventory from broker
        self.fastmcp_inventory = self.load_inventory()
        self.fastmcp_inventory.setdefault("host", "0.0.0.0")
        self.fastmcp_inventory.setdefault("port", 8001)

        # instantiate cache
        self.cache_dir = os.path.join(self.base_dir, "cache")
        os.makedirs(self.cache_dir, exist_ok=True)
        self.cache = self.get_diskcache()
        self.cache.expire()

        # start FastMCP server
        self.fastmcp_start()

        self.service_tasks_discovery_thread = threading.Thread(
            target=service_tasks_discovery, args=(self,)
        )
        self.service_tasks_discovery_thread.start()

        self.init_done_event.set()

    def get_diskcache(self) -> FanoutCache:
        """
        Initializes and returns a FanoutCache object.

        The FanoutCache is configured with the following parameters:

        - directory: The directory where the cache will be stored.
        - shards: Number of shards to use for the cache.
        - timeout: Timeout for cache operations in seconds.
        - size_limit: Maximum size of the cache in bytes.

        Returns:
            FanoutCache: An instance of FanoutCache configured with the specified parameters.
        """
        return FanoutCache(
            directory=self.cache_dir,
            shards=4,
            timeout=1,  # 1 second
            size_limit=1073741824,  #  1 GigaByte
        )

    def worker_exit(self) -> None:
        os.kill(os.getpid(), signal.SIGTERM)

    @Task(fastapi={"methods": ["GET"]})
    def get_version(self) -> Result:
        """
        Retrieves version information for key libraries and the current Python environment.

        Returns:
            Result: An object containing a dictionary with the version numbers of
                'norfab', 'mcp', 'uvicorn', 'pydantic', the Python version, and the platform.
                If a package is not found, its version will be an empty string.
        """

        libs = {
            "norfab": "",
            "uvicorn": "",
            "pydantic": "",
            "mcp": "",
            "python": sys.version.split(" ")[0],
            "platform": sys.platform,
        }
        # get version of packages installed
        for pkg in libs.keys():
            try:
                libs[pkg] = importlib.metadata.version(pkg)
            except importlib.metadata.PackageNotFoundError:
                pass

        return Result(task=f"{self.name}:get_version", result=libs)

    @Task(fastapi={"methods": ["GET"]})
    def get_inventory(self) -> Result:
        """
        Retrieves the current inventory from the FastMCP worker.

        Returns:
            Result: An object containing a copy of the worker's inventory and the task name.
        """
        return Result(
            result={**self.fastmcp_inventory},
            task=f"{self.name}:get_inventory",
        )

    @Task(fastapi={"methods": ["GET"]})
    def get_tools(
        self, brief: bool = False, service: str = "all", name: str = "*"
    ) -> Result:
        """
        Retrieve tools from the available norfab services, optionally filtered by service
        name and tool name pattern.

        Args:
            brief (bool, optional): If True, returns a list of tool names. If False,
                returns a dictionary with tool details.
            service (str, optional): The name of the service to filter tools by.
                Use "all" to include all services.
            name (str, optional): A glob pattern to match tool names.

        Returns:
            Result: An object containing the filtered tools. If brief is True, result
                is a list of tool names. Otherwise, result is a dictionary mapping tool
                names to their details.
        """
        ret = Result(
            result={},
            task=f"{self.name}:get_tools",
        )
        if brief:
            ret.result = []
            for service_name, tasks in self.norfab_services_tasks.items():
                if service == "all" or service_name == service:
                    for tool_name, tool_data in tasks.items():
                        if fnmatch(tool_name, name):
                            ret.result.append(tool_name)
        else:
            for service_name, tasks in self.norfab_services_tasks.items():
                if service == "all" or service_name == service:
                    for tool_name, tool_data in tasks.items():
                        if fnmatch(tool_name, name):
                            ret.result[tool_name] = tool_data["tool"].model_dump()

        return ret

    @Task(fastapi={"methods": ["POST"]})
    def discover(self, job, service: str = "all", progress: bool = True) -> Result:
        """
        Discovers available services tasks and auto-generate tools for them.

        Args:
            service (str, optional): The name of the service to discover. Defaults to "all".

        Returns:
            Result: An object containing the discovery results for the specified service.
        """
        job.event("Discovering NorFab services tasks")
        ret = Result(task=f"{self.name}:discover")
        ret.result = service_tasks_discovery(self, cycles=1, discover_service=service)

        return ret

    @Task(fastapi={"methods": ["GET"]})
    def get_status(self) -> Result:
        """
        Retrieves the current status of the application, including its name,
        URL, and the count of available tools.

        Returns:
            Result: An object containing a dictionary with the application's name,
                URL, and the number of tools, as well as the task identifier.
        """
        tools = self.get_tools(brief=True).result

        return Result(
            result={
                "name": self.app.name,
                "url": f"http://{self.fastmcp_inventory['host']}:{self.fastmcp_inventory['port']}/mcp/",
                "tools_count": len(tools),
            },
            task=f"{self.name}:get_status",
        )

    def fastmcp_start(self) -> None:
        """
        Starts the FastMCP server for the NorFab MCP application.

        This method initializes a FastMCP application instance with
        the specified host and port from `self.fastmcp_inventory`.

        It registers two MCP server endpoints:

          - `list_tools`: Asynchronously returns a list of available
            tools by aggregating all tools from `self.norfab_services_tasks`.
          - `call_tool`: Asynchronously handles tool invocation requests by
            parsing the tool name, extracting the corresponding service and
            task, and running the job using `self.client.run_job`.

        The FastMCP server is started in a separate thread using the
        "streamable-http" transport.
        """
        self.app = FastMCP(
            self.mcp_server_name,
            port=self.fastmcp_inventory["port"],
            host=self.fastmcp_inventory["host"],
        )

        @self.app._mcp_server.list_tools()
        async def list_tools() -> list[types.Tool]:
            ret = []
            for service, tasks in self.norfab_services_tasks.items():
                for tool_name, tool_data in tasks.items():
                    ret.append(tool_data["tool"])  # types.Tool object
            return ret

        @self.app._mcp_server.call_tool()
        async def call_tool(name: str, arguments: dict[str, Any]) -> dict[str, Any]:
            log.info(f"Calling tool '{name}' with arguments: '{arguments}'")

            # form NorFab service and task names
            service, tool_name = name.split("__")
            service = service[8:]
            tool_name = tool_name[5:]
            task_name = self.norfab_services_tasks[service][name]["task"]["name"]

            log.info(
                f"Calling NorFab service '{service}' task '{task_name}' with arguments: '{arguments}'"
            )

            return self.client.run_job(
                service=service,
                task=task_name,
                kwargs=arguments,
                workers="all",
            )

        self.app_server_thread = threading.Thread(
            target=self.app.run, kwargs={"transport": "streamable-http"}
        )
        self.app_server_thread.start()

        log.info(
            f"{self.name} - MCP server started, serving FastMCP app at "
            f"http://{self.fastmcp_inventory['host']}:{self.fastmcp_inventory['port']}/mcp/"
        )



import json
import builtins

from enum import Enum
from pydantic import (
    StrictBool,
    StrictStr,
    Field,
)
from ..common import ClientRunJobArgs, log_error_or_result, listen_events
from .nornir_picle_shell_common import (
    NorniHostsFilters,
    TabulateTableModel,
    NornirCommonArgs,
    print_nornir_results,
)
from typing import Union, Optional, Dict, List
from nornir_salt.plugins.functions import TabulateFormatter
from picle.models import PipeFunctionsModel, Outputters


class EnumTableTypes(str, Enum):
    table_brief = "brief"
    table_terse = "terse"
    table_extend = "extend"


class NornirTestShell(
    NorniHostsFilters, TabulateTableModel, NornirCommonArgs, ClientRunJobArgs
):
    suite: StrictStr = Field(..., description="Nornir suite nf://path/to/file.py")
    dry_run: Optional[StrictBool] = Field(
        None,
        description="Return produced per-host tests suite content without running tests",
        json_schema_extra={"presence": True},
        alias="dry-run",
    )
    subset: Optional[StrictStr] = Field(
        None,
        description="Filter tests by name",
    )
    failed_only: Optional[StrictBool] = Field(
        None,
        description="Return test results for failed tests only",
        json_schema_extra={"presence": True},
        alias="failed-only",
    )
    remove_tasks: Optional[StrictBool] = Field(
        None,
        description="Include/Exclude tested task results",
        json_schema_extra={"presence": True},
        alias="remove-tasks",
    )
    job_data: Optional[StrictStr] = Field(
        None, description="Path to YAML file with job data", alias="job-data"
    )
    table: Union[EnumTableTypes, Dict, StrictBool] = Field(
        "brief",
        description="Table format (brief, terse, extend) or parameters or True",
        json_schema_extra={"presence": "brief"},
    )
    groups: Union[StrictStr, List[StrictStr]] = Field(
        None,
        description="Test groups to run",
    )

    @staticmethod
    def source_suite():
        return ClientRunJobArgs.walk_norfab_files()

    @staticmethod
    def source_job_data():
        return ClientRunJobArgs.walk_norfab_files()

    @staticmethod
    @listen_events
    def run(uuid, *args, **kwargs):
        NFCLIENT = builtins.NFCLIENT
        workers = kwargs.pop("workers", "all")
        timeout = kwargs.pop("timeout", 600)
        verbose_result = kwargs.pop("verbose_result", False)
        nowait = kwargs.pop("nowait", False)
        dry_run = kwargs.get("dry_run", False)

        # extract job_data
        if kwargs.get("job_data") and not kwargs["job_data"].startswith("nf://"):
            kwargs["job_data"] = json.loads(kwargs["job_data"])

        if kwargs.get("groups") and isinstance(kwargs["groups"], str):
            kwargs["groups"] = [kwargs["groups"]]

        # extract Tabulate arguments
        table = kwargs.pop("table", {})  # tabulate
        headers = kwargs.pop("headers", "keys")  # tabulate
        headers_exclude = kwargs.pop("headers_exclude", [])  # tabulate
        sortby = kwargs.pop("sortby", "host")  # tabulate
        reverse = kwargs.pop("reverse", False)  # tabulate

        if table and not (verbose_result or dry_run):
            kwargs["add_details"] = True
            kwargs["to_dict"] = False

        result = NFCLIENT.run_job(
            "nornir",
            "test",
            workers=workers,
            args=args,
            kwargs=kwargs,
            uuid=uuid,
            timeout=timeout,
            nowait=nowait,
        )

        if nowait:
            return result, Outputters.outputter_nested

        result = log_error_or_result(result, verbose_result=verbose_result)

        # form table results
        if verbose_result or dry_run:
            ret = (
                result,
                Outputters.outputter_nested,
            )
        elif table:
            table_data = []
            for w_name, w_res in result.items():
                for item in w_res:
                    item["worker"] = w_name
                    table_data.append(item)
            ret = TabulateFormatter(
                table_data,
                tabulate=table,
                headers=headers,
                headers_exclude=headers_exclude,
                sortby=sortby,
                reverse=reverse,
            )
        else:
            ret = result

        return ret

    class PicleConfig:
        subshell = True
        prompt = "nf[nornir-test]#"
        outputter = print_nornir_results
        pipe = PipeFunctionsModel
